@@ -17,6 +17,10 @@
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 
+# Added for python2-3 compatibility
+from __future__ import print_function
+from __future__ import unicode_literals
+
 import struct
 from datetime import datetime
 
@@ -38,12 +42,14 @@ _global_warning_messages = []
 def warn(msg):
     if msg not in _global_warning_messages:
         _global_warning_messages.append(msg)
-        print "Warning: %s" % (msg)
+        print("Warning: %s" % (msg))
+
 
 def parse_windows_timestamp(qword):
     # see http://integriography.wordpress.com/2010/01/16/using-phython-to-parse-and-present-windows-64-bit-timestamps/
     return datetime.utcfromtimestamp(float(qword) * 1e-7 - 11644473600 )
 
+
 class RegistryException(Exception):
     """
     Base Exception class for Windows Registry access.
@@ -61,6 +67,7 @@ class RegistryException(Exception):
     def __str__(self):
         return "Registry Exception: %s" % (self._value)
 
+
 class RegistryStructureDoesNotExist(RegistryException):
     """
     Exception to be raised when a structure or block is requested which does not exist.
@@ -78,6 +85,7 @@ class RegistryStructureDoesNotExist(RegistryException):
     def __str__(self):
         return "Registry Structure Does Not Exist Exception: %s" % (self._value)
 
+
 class ParseException(RegistryException):
     """
     An exception to be thrown during Windows Registry parsing, such as
@@ -94,6 +102,7 @@ class ParseException(RegistryException):
     def __str__(self):
         return "Registry Parse Exception(%s)" % (self._value)
 
+
 class UnknownTypeException(RegistryException):
     """
     An exception to be raised when an unknown data type is encountered.
@@ -122,6 +131,7 @@ class UnknownTypeException(RegistryException):
     def __str__(self):
         return "Unknown Type Exception(%s)" % (self._value)
 
+
 class RegistryBlock(object):
     """
     Base class for structure blocks in the Windows Registry.
@@ -150,7 +160,7 @@ class RegistryBlock(object):
         Arguments:
         - `offset`: The relative offset from the start of the block.
         """
-        return struct.unpack_from("<H", self._buf, self._offset + offset)[0]
+        return struct.unpack_from(str("<H"), self._buf, self._offset + offset)[0]
 
     def unpack_dword(self, offset):
         """
@@ -158,7 +168,7 @@ class RegistryBlock(object):
         Arguments:
         - `offset`: The relative offset from the start of the block.
         """
-        return struct.unpack_from("<I", self._buf, self._offset + offset)[0]
+        return struct.unpack_from(str("<I"), self._buf, self._offset + offset)[0]
 
     def unpack_int(self, offset):
         """
@@ -166,7 +176,7 @@ class RegistryBlock(object):
         Arguments:
         - `offset`: The relative offset from the start of the block.
         """
-        return struct.unpack_from("<i", self._buf, self._offset + offset)[0]
+        return struct.unpack_from(str("<i"), self._buf, self._offset + offset)[0]
 
     def unpack_qword(self, offset):
         """
@@ -174,16 +184,16 @@ class RegistryBlock(object):
         Arguments:
         - `offset`: The relative offset from the start of the block.
         """
-        return struct.unpack_from("<Q", self._buf, self._offset + offset)[0]
+        return struct.unpack_from(str("<Q"), self._buf, self._offset + offset)[0]
 
     def unpack_string(self, offset, length):
         """
-        Returns a string from the relative offset with the given length.
+        Returns a byte string from the relative offset with the given length.
         Arguments:
         - `offset`: The relative offset from the start of the block.
         - `length`: The length of the string.
         """
-        return struct.unpack_from("<%ds" % (length), self._buf, self._offset + offset)[0]
+        return struct.unpack_from(str("<%ds") % (length), self._buf, self._offset + offset)[0]
 
     def absolute_offset(self, offset):
         """
@@ -205,6 +215,7 @@ class RegistryBlock(object):
         """
         return self._offset
 
+
 class REGFBlock(RegistryBlock):
     """
     The Windows Registry file header. This block has a length of 4k, although
@@ -260,7 +271,7 @@ class REGFBlock(RegistryBlock):
         return self.unpack_dword(0x28)
 
     def first_key(self):
-        first_hbin = self.hbins().next()
+        first_hbin = next(self.hbins())
 
         key_offset = first_hbin.absolute_offset(self.unpack_dword(0x24))
 
@@ -278,6 +289,7 @@ class REGFBlock(RegistryBlock):
             h = h.next()
             yield h
 
+
 class HBINCell(RegistryBlock):
     """
     HBIN data cell. An HBINBlock is continuously filled with HBINCell structures.
@@ -373,26 +385,27 @@ class HBINCell(RegistryBlock):
             raise RegistryStructureDoesNotExist("HBINCell is free at 0x%x" % (self.offset()))
 
         id_ = self.data_id()
-
-        if id_ == "vk":
+        
+        if id_ == b"vk":
             return VKRecord(self._buf, self.data_offset(), self)
-        elif id_ == "nk":
+        elif id_ == b"nk":
             return NKRecord(self._buf, self.data_offset(), self)
-        elif id_ == "lf":
+        elif id_ == b"lf":
             return LFRecord(self._buf, self.data_offset(), self)
-        elif id_ == "lh":
+        elif id_ == b"lh":
             return LHRecord(self._buf, self.data_offset(), self)
-        elif id_ == "li":
+        elif id_ == b"li":
             return LIRecord(self._buf, self.data_offset(), self)
-        elif id_ == "ri":
+        elif id_ == b"ri":
             return RIRecord(self._buf, self.data_offset(), self)
-        elif id_ == "sk":
+        elif id_ == b"sk":
             return SKRecord(self._buf, self.data_offset(), self)
-        elif id_ == "db":
+        elif id_ == b"db":
             return DBRecord(self._buf, self.data_offset(), self)
         else:
             return DataRecord(self._buf, self.data_offset(), self)
 
+
 class Record(RegistryBlock):
     """
     Abstract class for Records contained by cells in HBINs
@@ -419,6 +432,7 @@ class Record(RegistryBlock):
 
         return h.first_hbin().offset() + offset
 
+
 class DataRecord(Record):
     """
     A DataRecord is a HBINCell that does not contain any further structural data, but
@@ -438,6 +452,7 @@ class DataRecord(Record):
     def __str__(self):
         return "Data Record at 0x%x" % (self.offset())
 
+
 class DBIndirectBlock(Record):
     """
     The DBIndirect block is a list of offsets to DataRecords with data
@@ -472,6 +487,7 @@ class DBIndirectBlock(Record):
             length -= size
         return str(b)
 
+
 class DBRecord(Record):
     """
     A DBRecord is a large data block, which is not thoroughly documented.
@@ -488,7 +504,7 @@ class DBRecord(Record):
         super(DBRecord, self).__init__(buf, offset, parent)
 
         _id = self.unpack_string(0x0, 2)
-        if _id != "db":
+        if _id != b"db":
             raise ParseException("Invalid DB Record ID")
 
     def __str__(self):
@@ -504,6 +520,41 @@ class DBRecord(Record):
         dbi = DBIndirectBlock(self._buf, cell.data_offset(), cell)
         return dbi.large_data(length)
 
+
+def decode_utf16le(s):
+    """
+    decode_utf16le attempts to decode a bytestring as UTF-16LE.
+      If the string has an odd length, or some unexpected feature,
+      this function does its best to handle the data. It does not
+      catch any Unicode-related exceptions, such as UnicodeDecodeError,
+      so these should be handled by the caller.
+
+    @type s: str
+    @param s: a bytestring to pase
+    @rtype: unicode
+    @return: the unicode string decoded from `s`
+    @raises: this function does not attempt to catch any Unicode-related exception, so the caller should handle these.
+    """
+    if b"\x00\x00" in s:
+        index = s.index(b"\x00\x00")
+        if index > 2:
+            if s[index - 2] != b"\x00"[0]: #py2+3 
+                #  61 00 62 00 63 64 00 00
+                #                    ^  ^-- end of string
+                #                    +-- index
+                s = s[:index + 2]
+            else:
+                #  61 00 62 00 63 00 00 00
+                #                 ^     ^-- end of string
+                #                 +-- index
+                s = s[:index + 3]
+    if (len(s) % 2) != 0:
+        s = s + b"\x00"
+    s = s.decode("utf16")
+    s = s.partition('\x00')[0]
+    return s
+
+
 class VKRecord(Record):
     """
     The VKRecord holds one name-value pair.  The data may be one many types,
@@ -521,7 +572,7 @@ class VKRecord(Record):
         super(VKRecord, self).__init__(buf, offset, parent)
 
         _id = self.unpack_string(0x0, 2)
-        if _id != "vk":
+        if _id != b"vk":
             raise ParseException("Invalid VK Record ID")
 
     def data_type_str(self):
@@ -598,12 +649,13 @@ class VKRecord(Record):
     def name(self):
         """
         Get the name, if it exists. If not, the empty string is returned.
+        @return: ascii string containing the name
         """
         if not self.has_name():
             return ""
         else:
             name_length = self.unpack_word(0x2)
-            return self.unpack_string(0x14, name_length)
+            return self.unpack_string(0x14, name_length).decode("windows-1252")
 
     def data_type(self):
         """
@@ -613,7 +665,16 @@ class VKRecord(Record):
 
     def data_length(self):
         """
-        Get the length of this value data.
+        Get the length of this value data. This is the actual length of the data that should be parsed for the value.
+        """
+        size = self.unpack_dword(0x4)
+        if size > 0x80000000:
+            size -= 0x80000000
+        return size
+
+    def raw_data_length(self):
+        """
+        Get the literal length of this value data. Some interpretation may be required to make sense of the value.
         """
         return self.unpack_dword(0x4)
 
@@ -621,7 +682,7 @@ class VKRecord(Record):
         """
         Get the offset to the raw data associated with this value.
         """
-        if self.data_length() < 5 or self.data_length() >= 0x80000000:
+        if self.raw_data_length() < 5 or self.raw_data_length() >= 0x80000000:
             return self.absolute_offset(0x8)
         else:
             return self.abs_offset_from_hbin_offset(self.unpack_dword(0x8))
@@ -658,16 +719,16 @@ class VKRecord(Record):
           Not currently supported. TODO.
         """
         data_type = self.data_type()
-        data_length = self.data_length()
+        data_length = self.raw_data_length()
         data_offset = self.data_offset()
 
         if data_type == RegSZ or data_type == RegExpandSZ:
             if data_length >= 0x80000000:
                 # data is contained in the data_offset field
-                s = struct.unpack_from("<%ds" % (4), self._buf, data_offset)[0]
+                s = struct.unpack_from(str("<%ds") % (4), self._buf, data_offset)[0]
             elif 0x3fd8 < data_length < 0x80000000:
                 d = HBINCell(self._buf, data_offset, self)
-                if d.data_id() == "db":
+                if d.data_id() == b"db":
                     # this should always be the case
                     # but empirical testing does not confirm this
                     s = d.child().large_data(data_length)
@@ -675,20 +736,8 @@ class VKRecord(Record):
                     s = d.raw_data()[:data_length]
             else:
                 d = HBINCell(self._buf, data_offset, self)
-                s = struct.unpack_from("<%ds" % (data_length), self._buf, d.data_offset())[0]
-
-            try:
-                s = s.decode("utf16").encode("utf8").decode("utf8") # iron out the kinks by
-            except UnicodeDecodeError:                              # converting to and back to a Python str
-                try:
-                    s = s.decode("utf8").encode("utf8").decode("utf8")
-                except UnicodeDecodeError:
-                    try:
-                        s = s.decode("utf8", "replace").encode("utf8").decode("utf8")
-                    except:
-                        print "Well at this point you are screwed."
-                        raise
-            s = s.partition('\x00')[0]
+                s = struct.unpack_from(str("<%ds") % (data_length), self._buf, d.data_offset())[0]
+            s = decode_utf16le(s)
             return s
         elif data_type == RegBin or data_type == RegNone:
             if data_length >= 0x80000000:
@@ -696,7 +745,7 @@ class VKRecord(Record):
                 return self._buf[data_offset:data_offset + data_length]
             elif 0x3fd8 < data_length < 0x80000000:
                 d = HBINCell(self._buf, data_offset, self)
-                if d.data_id() == "db":
+                if d.data_id() == b"db":
                     # this should always be the case
                     # but empirical testing does not confirm this
                     return d.child().large_data(data_length)
@@ -712,7 +761,7 @@ class VKRecord(Record):
                 return []
             elif 0x3fd8 < data_length < 0x80000000:
                 d = HBINCell(self._buf, data_offset, self)
-                if d.data_id() == "db":
+                if d.data_id() == b"db":
                     s = d.child().large_data(data_length)
                 else:
                     s = d.raw_data()[:data_length]
@@ -722,27 +771,34 @@ class VKRecord(Record):
             return s.split("\x00")
         elif data_type == RegQWord:
             d = HBINCell(self._buf, data_offset, self)
-            return struct.unpack_from("<Q", self._buf, d.data_offset())[0]
+            return struct.unpack_from(str("<Q"), self._buf, d.data_offset())[0]
         elif data_type == RegBigEndian:
-            warn("Data type RegBigEndian not yet supported")
-            return False
-        elif data_type == RegLink:
-            warn("Data type RegLink not yet supported")
-            return False
-        elif data_type == RegResourceList:
-            warn("Data type RegResourceList not yet supported")
-            return False
-        elif data_type == RegFullResourceDescriptor:
-            warn("Data type RegFullResourceDescriptor not yet supported")
-            return False
-        elif data_type == RegResourceRequirementsList:
-            warn("Data type RegResourceRequirementsList not yet supported")
-            return False
+            d = HBINCell(self._buf, data_offset, self)
+            return struct.unpack_from(str(">I"), self._buf, d.data_offset())[0]
+        elif data_type == RegLink or \
+                        data_type == RegResourceList or \
+                        data_type == RegFullResourceDescriptor or \
+                        data_type == RegResourceRequirementsList:
+            # we don't really support these types, but can at least
+            #  return raw binary for someone else to work with.
+            if data_length >= 0x80000000:
+                data_length -= 0x80000000
+                return self._buf[data_offset:data_offset + data_length]
+            elif 0x3fd8 < data_length < 0x80000000:
+                d = HBINCell(self._buf, data_offset, self)
+                if d.data_id() == b"db":
+                    # this should always be the case
+                    # but empirical testing does not confirm this
+                    return d.child().large_data(data_length)
+                else:
+                    return d.raw_data()[:data_length]
+            return self._buf[data_offset + 4:data_offset + 4 + data_length]
         elif data_length < 5 or data_length >= 0x80000000:
             return self.unpack_dword(0x8)
         else:
             raise UnknownTypeException("Unknown VK Record type 0x%x at 0x%x" % (data_type, self.offset()))
 
+
 class SKRecord(Record):
     """
     Security Record. Contains Windows security descriptor,
@@ -762,7 +818,7 @@ class SKRecord(Record):
         super(SKRecord, self).__init__(buf, offset, parent)
 
         _id = self.unpack_string(0x0, 2)
-        if _id != "sk":
+        if _id != b"sk":
             raise ParseException("Invalid SK Record ID")
 
         self._offset_prev_sk = self.unpack_dword(0x4)
@@ -771,6 +827,7 @@ class SKRecord(Record):
     def __str__(self):
         return "SK Record at 0x%x" % (self.offset())
 
+
 class ValuesList(HBINCell):
     """
     A ValuesList is a simple structure of fixed length pointers/offsets to VKRecords.
@@ -803,6 +860,7 @@ class ValuesList(HBINCell):
             value_item += 4
             yield v
 
+
 class SubkeyList(Record):
     """
     A base class for use by structures recording the subkeys of Registry key.
@@ -832,6 +890,7 @@ class SubkeyList(Record):
         """
         return
 
+
 class RIRecord(SubkeyList):
     """
     The RIRecord is a structure linking to structures containing
@@ -870,6 +929,7 @@ class RIRecord(SubkeyList):
 
             key_index += 4
 
+
 class DirectSubkeyList(SubkeyList):
     def __init__(self, buf, offset, parent):
         """
@@ -898,6 +958,7 @@ class DirectSubkeyList(SubkeyList):
             yield NKRecord(self._buf, d.data_offset(), self)
             key_index += 8
 
+
 class LIRecord(DirectSubkeyList):
     """
     The LIRecord is a simple structure containing a list of offsets/pointers
@@ -930,6 +991,7 @@ class LIRecord(DirectSubkeyList):
             yield NKRecord(self._buf, d.data_offset(), self)
             key_index += 4
 
+
 class LFRecord(DirectSubkeyList):
     """
     The LFRecord is a simple structure containing a list of offsets/pointers
@@ -947,12 +1009,13 @@ class LFRecord(DirectSubkeyList):
         """
         super(LFRecord, self).__init__(buf, offset, parent)
         _id = self.unpack_string(0x0, 2)
-        if _id != "lf":
+        if _id != b"lf":
             raise ParseException("Invalid LF Record ID")
 
     def __str__(self):
         return "LFRecord(Length: %d) at 0x%x" % (self._keys_len(), self.offset())
 
+
 class LHRecord(DirectSubkeyList):
     """
     The LHRecord is a simple structure containing a list of offsets/pointers
@@ -971,12 +1034,13 @@ class LHRecord(DirectSubkeyList):
         """
         super(LHRecord, self).__init__(buf, offset, parent)
         _id = self.unpack_string(0x0, 2)
-        if _id != "lh":
+        if _id != b"lh":
             raise ParseException("Invalid LH Record ID")
 
     def __str__(self):
         return "LHRecord(Length: %d) at 0x%x" % (self._keys_len(), self.offset())
 
+
 class NKRecord(Record):
     """
     The NKRecord defines the tree-like structure of the Windows Registry.
@@ -993,7 +1057,7 @@ class NKRecord(Record):
         """
         super(NKRecord, self).__init__(buf, offset, parent)
         _id = self.unpack_string(0x0, 2)
-        if _id != "nk":
+        if _id != b"nk":
             raise ParseException("Invalid NK Record ID")
 
     def __str__(self):
@@ -1019,6 +1083,7 @@ class NKRecord(Record):
     def classname(self):
         """
         If this has a classname, get it as a string. Otherwise, return the empty string.
+        @return: unicode string containg the class name
         """
         if not self.has_classname():
             return ""
@@ -1028,7 +1093,7 @@ class NKRecord(Record):
 
         offset = self.abs_offset_from_hbin_offset(classname_offset)
         d = HBINCell(self._buf, offset, self)
-        return struct.unpack_from("<%ds" % (classname_length), self._buf, d.data_offset())[0].decode("utf-16le").rstrip("\x00")
+        return struct.unpack_from(str("<%ds") % (classname_length), self._buf, d.data_offset())[0].decode("utf-16le").rstrip("\x00")
 
     def timestamp(self):
         """
@@ -1039,13 +1104,15 @@ class NKRecord(Record):
     def name(self):
         """
         Return the registry key name as a string.
+        @return: ascii string containing the name
         """
         name_length = self.unpack_word(0x48)
-        return self.unpack_string(0x4C, name_length)
+        return self.unpack_string(0x4C, name_length).decode("windows-1252")
 
     def path(self):
         """
-        Return the full path of the registry key as a string.
+        Return the full path of the registry key as a unicode string
+        @return: unicode string containing the path
         """
         name = ""
         p = self
@@ -1140,19 +1207,20 @@ class NKRecord(Record):
         d = HBINCell(self._buf, subkey_list_offset, self)
         id_ = d.data_id()
 
-        if id_ == "lf":
+        if id_ == b"lf":
             l = LFRecord(self._buf, d.data_offset(), self)
-        elif id_ == "lh":
+        elif id_ == b"lh":
             l = LHRecord(self._buf, d.data_offset(), self)
-        elif id_ == "ri":
+        elif id_ == b"ri":
             l = RIRecord(self._buf, d.data_offset(), self)
-        elif id_ == "li":
+        elif id_ == b"li":
             l = LIRecord(self._buf, d.data_offset(), self)
         else:
             raise ParseException("Subkey list with type %s encountered, but not yet supported." % (id_))
 
         return l
 
+
 class HBINBlock(RegistryBlock):
     """
     An HBINBlock is the basic allocation block of the Windows Registry.
